from io import StringIO
from typing import Union, Tuple

from sourced.ml.utils import PickleableLogger  # nopep8


class Transformer(PickleableLogger):
    BLOCKS = False  # set to True if __call__ launches PySpark

    def __init__(self, explain=None, **kwargs):
        super().__init__(**kwargs)
        self._children = []
        self._parent = None
        self._explained = explain

    def __getstate__(self):
        state = super().__getstate__()
        del state["_parent"]
        del state["_children"]
        return state

    @property
    def explained(self):
        if self._explained is None and self.parent is not None:
            return self.parent.explained
        return bool(self._explained)

    @property
    def children(self):
        return tuple(self._children)

    @property
    def parent(self):
        return self._parent

    def path(self):
        node = self
        path = []
        while node is not None:
            path.append(node)
            node = node.parent
        path.reverse()
        return path

    def link(self, *transformer: "Transformer") -> Union[Tuple["Transformer"], "Transformer"]:
        def link_one(t: Transformer) -> Transformer:
            self._children.append(t)
            t._parent = self
            return t

        if len(transformer) == 1:
            return link_one(transformer[0])
        return tuple(link_one(t) for t in transformer)

    def unlink(self, *transformer: "Transformer"):
        for t in transformer:
            self._children.remove(t)
            t._parent = None
        return self

    def __rshift__(self, other):
        """Shortcut for link"""
        if isinstance(other, (list, tuple)):
            return self.link(*other)
        return self.link(other)

    def __lshift__(self, other):
        """Shortcut for unlink"""
        if isinstance(other, (list, tuple)):
            return self.unlink(*other)
        return self.unlink(other)

    def _explode(self, head, context):
        if context[-1] is not self:
            context.append(self)
            if not self._children or self.BLOCKS:
                self._log.info(self._format_pipeline(context))
            head = self(head)
        results = []
        for child in self._children:
            results.extend(child._explode(head, context.copy()))
        else:
            results.append(head)
        return results

    def explode(self, head=None):
        """
        Execute all the branches in the tree going from the node. The node itself
        is execute()-ed.
        :param head: The input to feed to the root. Can be None - the default one \
                     will be used if possible.
        :return: The results from all the leaves.
        """
        head = self.execute(head)
        pipeline = [self]
        node = self
        while node.parent is not None:
            node = node.parent
            pipeline.append(node)
        pipeline.reverse()
        return self._explode(head, pipeline)

    def execute(self, head=None):
        """
        Execute the node together with all its dependencies, in order.
        :param head: The input to feed to the ultimate parent. Can be None - the default one \
                     will be used if possible.
        :return: The result of the execution.
        """
        pipeline = self.path()
        if not self._children:
            self._log.info(self._format_pipeline(pipeline))
        for node in pipeline:
            head = node(head)
        return head

    def graph(self, name="source-d", stream=None):
        if stream is None:
            stream = StringIO()
        stream.write("digraph %s {\n" % name)
        counters = {}
        nodes = {}
        queue = [(self, None)]
        while queue:
            node, parent = queue.pop(0)
            try:
                myself = nodes[node]
            except KeyError:
                index = counters.setdefault(type(node), 0)
                index += 1
                counters[type(node)] = index
                myself = "%s %d" % (type(node).__name__, index)
                nodes[node] = myself
            if parent is not None:
                stream.write("\t\"%s\" -> \"%s\"\n" % (parent, myself))
            for child in node._children:
                queue.append((child, myself))
        stream.write("}\n")
        return stream

    def _get_log_name(self):
        return self.__class__.__name__

    @staticmethod
    def _format_pipeline(pipeline):
        return " -> ".join(type(n).__name__ for n in pipeline)

    def __call__(self, head):
        raise NotImplementedError()


class Execute(Transformer):
    """
    Special transformer to execute all the pipeline.
    As soon as one links anything to this Transformer it call execute() for the pipeline.
    It is not possible to link anything to this transformer.
    """

    def __init__(self, head=None, explain=None, **kwargs):
        super().__init__(explain, **kwargs)
        self.head = head
        self._real_parent = None

    @property
    def _parent(self):
        return self._real_parent

    @_parent.setter
    def _parent(self, value: Transformer):
        self._real_parent = value
        if value is not None:
            value.execute(self.head)

    def link(self, *transformer: "Transformer") -> Union[Tuple["Transformer"], "Transformer"]:
        raise AssertionError("It is not possible to link anything after Leaf.")

    def __call__(self, head):
        return head



from typing import Union, Dict
from pyspark.rdd import RDD
from pyspark import Row

from sourced.ml.transformers.transformer import Transformer
from sourced.ml.models import DocumentFrequencies


class Indexer(Transformer):
    """
    Maps each value of the given column in an RDD of pyspark.sql.Row to the respective integer
    index. The mapping is created by collecting all the unique values, sorting them and finally
    enumerating. Use value_to_index or [] to get index value.
    """
    def __init__(self, column: Union[int, str], column2id: Dict[str, int]=None, **kwargs):
        """
        :param column: column index or its name in pyspark.RDD for indexing.
        :param column2id: precomputed mapping between column values to integers.
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.column = column
        self._value_to_index = column2id
        self._values = None

    def __getitem__(self, key: Union[int, str]):
        """
        Get index for given value or value for given index.
        If key type is string, __getitem__ gives you index value for key.
        If key type is int, __getitem__ gives you value for this index.
        WARNING: Do not use this function if you index non-string values, use `value_to_index` and
        values properties instead.

        :param key: value or index.
        :return: index for given value or value for given index.
        """
        if not isinstance(key, (int, str)):
            raise TypeError("__getitem__ supports only for string and int types. Use "
                            "`value_to_index` property to get index value.")
        return self.value_to_index[key] if isinstance(key, str) else self.values[key]

    def __len__(self):
        return len(self._value_to_index)

    @property
    def value_to_index(self):
        if self._value_to_index is None:
            raise AttributeError("column2id value not available. Run Indexer first.")
        return self._value_to_index

    def values(self):
        if self._values is None:
            self._values = [None] * len(self)
            for k, v in self.value_to_index.items():
                self._values[v] = k
        return self._values

    def calculate_value_to_index(self, rdd: RDD):
        column_name = self.column
        if isinstance(column_name, str):
            column = rdd.map(lambda x: getattr(x, column_name))
        else:
            column = rdd.map(lambda x: x[column_name])
        self._log.info("Collecting the list of distinct sorted values (%s)", column_name)
        values = column.distinct()
        if self.explained:
            self._log.info("toDebugString():\n%s", values.toDebugString().decode())
        values = values.collect()
        values.sort()  # We do not expect an extraordinary number of distinct values
        self._log.info("%d distinct values", len(values))
        if len(values) == 0:
            raise RuntimeError("Number of distinct values is zero.")
        self._value_to_index = {d: i for i, d in enumerate(values)}

    def save_index(self, path):
        self._log.info("Saving the index to %s", path)
        DocumentFrequencies().construct(len(self._value_to_index), self._value_to_index) \
            .save(path)

    def __call__(self, rdd: RDD):
        column_name = self.column
        if self._value_to_index is None:
            self.calculate_value_to_index(rdd)
        column2id = rdd.context.broadcast(self._value_to_index)

        def index_column(row):
            """
            Map column_id column to its index value stored in column2id.
            WARNING: due to pyspark documentation
            (http://spark.apache.org/docs/latest/rdd-programming-guide.html#passing-functions-to-spark)
            do not use self inside this function. It will be suboptimal and probably fail to run.
            Please contact me if you have troubles: kslavnov@gmail.com
            """
            if isinstance(column_name, str):
                assert isinstance(row, Row)
                row_dict = row.asDict()
                row_dict[column_name] = column2id.value[row_dict[column_name]]
                return Row(**row_dict)
            return row[:column_name] + (column2id.value[row[column_name]],) + row[column_name + 1:]
        indexed_rdd = rdd.map(index_column)
        column2id.unpersist(blocking=True)
        return indexed_rdd



from sourced.ml.algorithms import UastRandomWalk2Bag
from sourced.ml.extractors import BagsExtractor, register_extractor, get_names_from_kwargs,\
    filter_kwargs


@register_extractor
class UastRandomWalkBagExtractor(BagsExtractor):
    NAME = "node2vec"
    NAMESPACE = "r."
    OPTS = dict(get_names_from_kwargs(UastRandomWalk2Bag.__init__))
    OPTS.update(BagsExtractor.OPTS)

    def __init__(self, docfreq_threshold=None, **kwargs):
        original_kwargs = kwargs
        uast2bag_kwargs = filter_kwargs(kwargs, UastRandomWalk2Bag.__init__)
        for k in uast2bag_kwargs:
            kwargs.pop(k)
        super().__init__(docfreq_threshold, **kwargs)
        self._log.debug("__init__ %s", original_kwargs)
        self.uast2bag = UastRandomWalk2Bag(**uast2bag_kwargs)

    def uast_to_bag(self, uast):
        return self.uast2bag(uast)



import re

import Stemmer


class TokenParser:
    """
    Common utilities for splitting and stemming tokens.
    """
    NAME_BREAKUP_RE = re.compile(r"[^a-zA-Z]+")  #: Regexp to split source code identifiers.
    STEM_THRESHOLD = 6  #: We do not stem splitted parts shorter than or equal to this size.
    MAX_TOKEN_LENGTH = 256  #: We cut identifiers longer than this value.
    MIN_SPLIT_LENGTH = 3  #: We do not split source code identifiers shorter than this value.
    DEFAULT_SINGLE_SHOT = False  #: True if we do not want to join small identifiers to next one.
    # Example: 'sourced.ml.algorithms' -> ["sourc", "sourcedml", "algorithm", "mlalgorithm"].
    # if True we have only ["sourc", "algorithm"].
    # if you do not want to filter small tokens set min_split_length=1.

    def __init__(self, stem_threshold=STEM_THRESHOLD, max_token_length=MAX_TOKEN_LENGTH,
                 min_split_length=MIN_SPLIT_LENGTH, single_shot=DEFAULT_SINGLE_SHOT):
        self._stemmer = Stemmer.Stemmer("english")
        self._stemmer.maxCacheSize = 0
        self._stem_threshold = stem_threshold
        self._max_token_length = max_token_length
        self._min_split_length = min_split_length
        self._single_shot = single_shot

    @property
    def stem_threshold(self):
        return self._stem_threshold

    @stem_threshold.setter
    def stem_threshold(self, value):
        if not isinstance(value, int):
            raise TypeError("stem_threshold must be an integer - got %s" % type(value))
        if value < 1:
            raise ValueError("stem_threshold must be greater than 0 - got %d" % value)
        self._stem_threshold = value

    @property
    def max_token_length(self):
        return self._max_token_length

    @max_token_length.setter
    def max_token_length(self, value):
        if not isinstance(value, int):
            raise TypeError("max_token_length must be an integer - got %s" % type(value))
        if value < 1:
            raise ValueError("max_token_length must be greater than 0 - got %d" % value)
        self._max_token_length = value

    @property
    def min_split_length(self):
        return self._min_split_length

    @min_split_length.setter
    def min_split_length(self, value):
        if not isinstance(value, int):
            raise TypeError("min_split_length must be an integer - got %s" % type(value))
        if value < 1:
            raise ValueError("min_split_length must be greater than 0 - got %d" % value)
        self._min_split_length = value

    def __call__(self, token):
        return self.process_token(token)

    def process_token(self, token):
        for word in self.split(token):
            yield self.stem(word)

    def stem(self, word):
        if len(word) <= self.stem_threshold:
            return word
        return self._stemmer.stemWord(word)

    def split(self, token):
        token = token.strip()[:self.max_token_length]

        def ret(name):
            r = name.lower()
            if len(name) >= self.min_split_length:
                ret.last_subtoken = r
                yield r
                if ret.prev_p and not self._single_shot:
                    yield ret.prev_p + r
                    ret.prev_p = ""
            elif not self._single_shot:
                    ret.prev_p = r
                    yield ret.last_subtoken + r
                    ret.last_subtoken = ""
        ret.prev_p = ""
        ret.last_subtoken = ""

        for part in self.NAME_BREAKUP_RE.split(token):
            if not part:
                continue
            prev = part[0]
            pos = 0
            for i in range(1, len(part)):
                this = part[i]
                if prev.islower() and this.isupper():
                    yield from ret(part[pos:i])
                    pos = i
                elif prev.isupper() and this.islower():
                    if 0 < i - 1 - pos <= self.min_split_length:
                        yield from ret(part[pos:i])
                        pos = i
                    elif i - 1 > pos:
                        yield from ret(part[pos:i])
                        pos = i
                prev = this
            last = part[pos:]
            if last:
                yield from ret(last)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_stemmer"]
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._stemmer = Stemmer.Stemmer("english")


class NoopTokenParser:
    """
    One can use this class if he or she does not want to do any parsing.
    """

    def process_token(self, token):
        yield token

    def __call__(self, token):
        return self.process_token(token)



import argparse
import json
import logging
from typing import Optional, Union, Iterable
import sys


from sourced.ml import extractors
from sourced.ml.transformers import BOWWriter, Moder
from sourced.ml.utils import add_engine_args


class ArgumentDefaultsHelpFormatterNoNone(argparse.ArgumentDefaultsHelpFormatter):
    """
    Pretty formatter of help message for arguments.
    It adds default value to the end if it is not None.
    """
    def _get_help_string(self, action):
        if action.default is None:
            return action.help
        return super()._get_help_string(action)


def handle_input_arg(input_arg: Union[str, Iterable[str]],
                     log: Optional[logging.Logger] = None):
    """
    Process input arguments and return an iterator over input files.

    :param input_arg: list of files to process or `-` to get \
        file paths from stdin.
    :param log: Logger if you want to log handling process.
    :return: An iterator over input files.
    """
    log = log.info if log else (lambda *x: None)
    if input_arg == "-" or input_arg == ['-']:
        log("Reading file paths from stdin.")
        for line in sys.stdin:
            yield line.strip()
    else:
        if isinstance(input_arg, str):
            yield input_arg
        else:
            yield from input_arg


def add_repartitioner_arg(my_parser: argparse.ArgumentParser):
    my_parser.add_argument(
        "--partitions", required=False, default=None, type=int,
        help="Performs data repartition to specified number of partitions. "
             "Nothing happens if parameter is unset.")
    my_parser.add_argument(
        "--shuffle", action="store_true",
        help="Use RDD.repartition() instead of RDD.coalesce().")


def add_split_stem_arg(my_parser: argparse.ArgumentParser):
    my_parser.add_argument(
        "--split", action="store_true",
        help="Split identifiers based on special characters or case changes. For example split "
             "'ThisIs_token' to ['this', 'is', 'token'].")


def add_vocabulary_size_arg(my_parser: argparse.ArgumentParser):
    my_parser.add_argument(
        "-v", "--vocabulary-size", default=10000000, type=int,
        help="The maximum vocabulary size.")


def add_min_docfreq(my_parser: argparse.ArgumentParser):
    my_parser.add_argument(
        "--min-docfreq", default=1, type=int,
        help="The minimum document frequency of each feature.")


def add_repo2_args(my_parser: argparse.ArgumentParser, default_packages=None):
    my_parser.add_argument(
        "-r", "--repositories", required=True,
        help="The path to the repositories.")
    my_parser.add_argument(
        "--parquet", action="store_true", help="Use Parquet files as input.")
    my_parser.add_argument(
        "--graph", help="Write the tree in Graphviz format to this file.")
    # TODO(zurk): get languages from bblfsh directly as soon as
    # https://github.com/bblfsh/client-scala/issues/98 resolved
    languages = ["Java", "Python", "Go", "JavaScript", "TypeScript", "Ruby", "Bash", "Php"]
    my_parser.add_argument(
        "-l", "--languages", nargs="+", choices=languages,
        default=None,  # Default value for --languages arg should be None.
                       # Otherwise if you process parquet files without 'lang' column, you will
                       # fail to process it with any --languages argument.
        help="The programming languages to analyse.")
    my_parser.add_argument("--blacklist", action="store_true",
                           help="Exclude the languages in --languages from the analysis "
                                "instead of filtering by default.")
    add_dzhigurda_arg(my_parser)
    add_engine_args(my_parser, default_packages)


def add_df_args(my_parser: argparse.ArgumentParser, required=True):
    my_parser.add_argument(
        "--min-docfreq", default=1, type=int,
        help="The minimum document frequency of each feature.")
    df_group = my_parser.add_mutually_exclusive_group(required=required)
    df_group.add_argument(
        "--docfreq-out", help="Path to save generated DocumentFrequencies model.")
    df_group.add_argument(
        "--docfreq-in", help="Path to load pre-generated DocumentFrequencies model.")
    add_vocabulary_size_arg(my_parser)


def add_feature_args(my_parser: argparse.ArgumentParser, required=True):
    my_parser.add_argument("-x", "--mode", choices=Moder.Options.__all__,
                           default="file", help="What to select for analysis.")
    my_parser.add_argument(
        "--quant", help="[IN/OUT] The path to the QuantizationLevels model.")
    my_parser.add_argument(
        "-f", "--feature", nargs="+",
        choices=[ex.NAME for ex in extractors.__extractors__.values()],
        required=required, help="The feature extraction scheme to apply.")
    for ex in extractors.__extractors__.values():
        for opt, val in ex.OPTS.items():
            my_parser.add_argument(
                "--%s-%s" % (ex.NAME, opt), default=val, type=json.loads,
                help="%s's kwarg" % ex.__name__)


def add_bow_args(my_parser: argparse.ArgumentParser):
    my_parser.add_argument(
        "--bow", required=True, help="[OUT] The path to the Bag-Of-Words model.")
    my_parser.add_argument(
        "--batch", default=BOWWriter.DEFAULT_CHUNK_SIZE, type=int,
        help="The maximum size of a single BOW file in bytes.")
    my_parser.add_argument(
        "--num-iterations", default=1, type=int,
        help="After partitioning by document we run the pipeline on each partition separately "
             "in a loop. This number indicates the number of partitions.")


def add_cached_index_arg(my_parser: argparse.ArgumentParser, create: bool = False):
    direction = "OUT" if create else "IN"
    my_parser.add_argument(
        "--cached-index-path", default=None, required=True,
        help="[%s] Path to the docfreq model holding the document's index." % direction)


def add_dzhigurda_arg(my_parser):
    my_parser.add_argument(
        "--dzhigurda", default=0, type=int,
        help="Number of the additional commits look over in the history starting from the HEAD "
             "commits. 0 corresponds to HEAD only commits, 1 to HEAD and HEAD~1, 2 to HEAD, HEAD~1"
             " and HEAD~2, etc. With `--dzhigurda -1` we keep all possible commits for each "
             "document.")



from pyspark import RDD, Row

from sourced.ml.extractors import Extractor
from sourced.ml.transformers.transformer import Transformer
from sourced.ml.utils import EngineConstants


class UastRow2Document(Transformer):
    REPO_PATH_SEP = "//"
    PATH_BLOB_SEP = "@"

    def __call__(self, rows: RDD) -> RDD:
        assert isinstance(rows, RDD)
        return rows.map(self.documentize)

    def documentize(self, r: Row) -> Row:
        ec = EngineConstants.Columns
        doc = r[ec.RepositoryId]
        if r[ec.Path]:
            doc += self.REPO_PATH_SEP + r[ec.Path]
        if r[ec.BlobId]:
            doc += self.PATH_BLOB_SEP + r[ec.BlobId]
        bfc = Uast2BagFeatures.Columns
        return Row(**{bfc.document: doc, ec.Uast: r[ec.Uast]})


class UastMiner(Transformer):
    def __init__(self, *extractors: Extractor, **kwargs):
        super().__init__(**kwargs)
        self.extractors = extractors

    def __call__(self, rows: RDD):
        return rows.flatMap(self.process_row)

    def process_row(self, row: Row):
        for uast in row[EngineConstants.Columns.Uast]:
            for extractor in self.extractors:
                for feature in extractor.extract(uast):
                    yield self.process_feature(row, extractor.NAME, feature)

    def process_feature(self, row: Row, name, feature):
        raise NotImplemented


class Uast2Features(UastMiner):
    def process_feature(self, row: Row, name, feature):
        new = row.asDict()
        new[name] = feature
        return new


class Uast2BagFeatures(UastMiner):
    class Columns:
        """
        Standard column names for interop.
        """
        token = "token"
        document = "document"
        value = "value"

    def process_feature(self, row: Row, name, feature):
        return (feature[0], row[self.Columns.document]), feature[1]



# flake8: noqa
from sourced.ml.transformers.basic import Sampler, Collector, First, Identity, Cacher, Ignition, \
    HeadFiles, UastExtractor, FieldsSelector, ParquetSaver, ParquetLoader, Repartitioner, \
    UastDeserializer, Counter, create_file_source, create_uast_source, CsvSaver, LanguageSelector,\
    DzhigurdaFiles, PartitionSelector, Distinct, LanguageExtractor, Rower
from sourced.ml.transformers.indexer import Indexer
from sourced.ml.transformers.tfidf import TFIDF
from sourced.ml.transformers.transformer import Transformer, Execute
from sourced.ml.transformers.uast2bag_features import Uast2Features, Uast2BagFeatures, UastRow2Document
from sourced.ml.transformers.uast2quant import Uast2Quant
from sourced.ml.transformers.bag_features2docfreq import BagFeatures2DocFreq
from sourced.ml.transformers.bag_features2termfreq import BagFeatures2TermFreq
from sourced.ml.transformers.content2ids import ContentToIdentifiers, IdentifiersToDataset
from sourced.ml.transformers.coocc import CooccConstructor, CooccModelSaver
from sourced.ml.transformers.bow_writer import BOWWriter
from sourced.ml.transformers.moder import Moder



#!/usr/bin/env python3
#
# Copyright 2016 Google Inc. All Rights Reserved.
# Copyright 2017 Sourced Technologies S. L.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Submatrix-wise Vector Embedding Learner.

Implementation of SwiVel algorithm described at:
http://arxiv.org/abs/1602.02215

This program expects an input directory that contains the following files.

  row_vocab.txt, col_vocab.txt

    The row an column vocabulary files.  Each file should contain one token per
    line; these will be used to generate a tab-separate file containing the
    trained embeddings.

  row_sums.txt, col_sum.txt

    The matrix row and column marginal sums.  Each file should contain one
    decimal floating point number per line which corresponds to the marginal
    count of the matrix for that row or column.

  shards.recs

    A file containing the sub-matrix shards, stored as TFRecords. Each shard is
    expected to be a serialzed tf.Example protocol buffer with the following
    properties:

      global_row: the global row indicies contained in the shard
      global_col: the global column indicies contained in the shard
      sparse_local_row, sparse_local_col, sparse_value: three parallel arrays
      that are a sparse representation of the submatrix counts.

It will generate embeddings, training from the input directory for
the specified number of epochs.  When complete, it will output the trained
vectors to a tab-separated file that contains one line per embedding.  Row and
column embeddings are stored in separate files.

"""

import glob
import math
import os
import time
import threading

import numpy as np
import tensorflow as tf
from tensorflow.python.client import device_lib

flags = tf.app.flags

flags.DEFINE_string("input_base_path", None,
                    "Directory containing input shards, vocabularies, "
                    "and marginals.")
flags.DEFINE_string("output_base_path", None,
                    "Path where to write the trained embeddings.")
flags.DEFINE_integer("embedding_size", 300, "Size of the embeddings")
flags.DEFINE_boolean("trainable_bias", False, "Biases are trainable")
flags.DEFINE_integer("submatrix_rows", 4096,
                     "Rows in each training submatrix. This must match "
                     "the training data.")
flags.DEFINE_integer("submatrix_cols", 4096,
                     "Rows in each training submatrix. This must match "
                     "the training data.")
flags.DEFINE_float("loss_multiplier", 1.0 / 4096,
                   "constant multiplier on loss.")
flags.DEFINE_float("confidence_exponent", 0.5,
                   "Exponent for l2 confidence function")
flags.DEFINE_float("confidence_scale", 0.25,
                   "Scale for l2 confidence function")
flags.DEFINE_float("confidence_base", 0.1, "Base for l2 confidence function")
flags.DEFINE_float("learning_rate", 1.0, "Initial learning rate")
flags.DEFINE_string("optimizer", "Adagrad",
                    "SGD optimizer (tf.train.*Optimizer)")
flags.DEFINE_integer("num_concurrent_steps", 2,
                     "Number of threads to train with")
flags.DEFINE_integer("num_readers", 4,
                     "Number of threads to read the input data and feed it")
flags.DEFINE_float("num_epochs", 40, "Number epochs to train for")
flags.DEFINE_float("per_process_gpu_memory_fraction", 0,
                   "Fraction of GPU memory to use, 0 means allow_growth")
flags.DEFINE_integer("num_gpus", 0,
                     "Number of GPUs to use, 0 means all available")
flags.DEFINE_string("logs", "",
                    "Path for TensorBoard logs (empty value disables them)")

FLAGS = flags.FLAGS


def log(message, *args, **kwargs):
    tf.logging.info(message, *args, **kwargs)


def get_available_gpus():
    return [d.name for d in device_lib.list_local_devices()
            if d.device_type == "GPU"]


def embeddings_with_init(vocab_size, embedding_dim, name):
    """Creates and initializes the embedding tensors."""
    return tf.get_variable(name=name,
                           shape=[vocab_size, embedding_dim],
                           initializer=tf.random_normal_initializer(
                               stddev=math.sqrt(1.0 / embedding_dim)))


def count_matrix_input(filenames, submatrix_rows, submatrix_cols):
    """Reads submatrix shards from disk."""
    filename_queue = tf.train.string_input_producer(filenames)
    reader = tf.WholeFileReader()
    _, serialized_example = reader.read(filename_queue)
    features = tf.parse_single_example(
        serialized_example,
        features={
            "global_row": tf.FixedLenFeature([submatrix_rows], dtype=tf.int64),
            "global_col": tf.FixedLenFeature([submatrix_cols], dtype=tf.int64),
            "sparse_local_row": tf.VarLenFeature(dtype=tf.int64),
            "sparse_local_col": tf.VarLenFeature(dtype=tf.int64),
            "sparse_value": tf.VarLenFeature(dtype=tf.float32)
        })

    global_row = features["global_row"]
    global_col = features["global_col"]

    sparse_local_row = features["sparse_local_row"].values
    sparse_local_col = features["sparse_local_col"].values
    sparse_count = features["sparse_value"].values

    sparse_indices = tf.concat(axis=1, values=[tf.expand_dims(sparse_local_row, 1),
                                               tf.expand_dims(sparse_local_col, 1)])
    count = tf.sparse_to_dense(sparse_indices, [submatrix_rows, submatrix_cols],
                               sparse_count, validate_indices=False)

    queued_global_row, queued_global_col, queued_count = tf.train.batch(
        [global_row, global_col, count],
        batch_size=1,
        num_threads=FLAGS.num_readers,
        capacity=32)

    queued_global_row = tf.reshape(queued_global_row, [submatrix_rows])
    queued_global_col = tf.reshape(queued_global_col, [submatrix_cols])
    queued_count = tf.reshape(queued_count, [submatrix_rows, submatrix_cols])

    return queued_global_row, queued_global_col, queued_count


def read_marginals_file(filename):
    """Reads text file with one number per line to an array."""
    with open(filename) as lines:
        return [float(line) for line in lines]


def write_embedding_tensor_to_disk(vocab_path, output_path, sess, embedding):
    """Writes tensor to output_path as tsv"""
    # Fetch the embedding values from the model
    embeddings = sess.run(embedding)

    with open(output_path, "w") as out_f:
        with open(vocab_path) as vocab_f:
            for index, word in enumerate(vocab_f):
                word = word.strip()
                embedding = embeddings[index]
                out_f.write(word + "\t" + "\t".join(
                    [str(x) for x in embedding]) + "\n")


def write_embeddings_to_disk(config, model, sess):
    """Writes row and column embeddings disk"""
    # Row Embedding
    row_vocab_path = config.input_base_path + "/row_vocab.txt"
    row_embedding_output_path = config.output_base_path + "/row_embedding.tsv"
    log("Writing row embeddings to: %s", row_embedding_output_path)
    write_embedding_tensor_to_disk(row_vocab_path, row_embedding_output_path,
                                   sess, model.row_embedding)

    # Column Embedding
    col_vocab_path = config.input_base_path + "/col_vocab.txt"
    col_embedding_output_path = config.output_base_path + "/col_embedding.tsv"
    log("Writing column embeddings to: %s", col_embedding_output_path)
    write_embedding_tensor_to_disk(col_vocab_path, col_embedding_output_path,
                                   sess, model.col_embedding)


class SwivelModel:
    """Small class to gather needed pieces from a Graph being built."""

    def __init__(self, config):
        """Construct graph for dmc."""
        self._config = config

        # Create paths to input data files
        log("Reading model from: %s", config.input_base_path)
        count_matrix_files = glob.glob(os.path.join(config.input_base_path, "shard-*.pb"))
        row_sums_path = os.path.join(config.input_base_path, "row_sums.txt")
        col_sums_path = os.path.join(config.input_base_path, "col_sums.txt")

        # Read marginals
        row_sums = read_marginals_file(row_sums_path)
        col_sums = read_marginals_file(col_sums_path)

        self.n_rows = len(row_sums)
        self.n_cols = len(col_sums)
        log("Matrix dim: (%d,%d) SubMatrix dim: (%d,%d)",
            self.n_rows, self.n_cols, config.submatrix_rows,
            config.submatrix_cols)
        if self.n_cols < config.submatrix_cols:
            raise ValueError(
                "submatrix_cols={0} can not be bigger than columns number={1} "
                "(specify submatrix_cols={1})".format(config.submatrix_cols, self.n_cols))
        if self.n_rows < config.submatrix_rows:
            raise ValueError(
                "submatrix_rows={0} can not be bigger than rows number={1} "
                "(specify submatrix_rows={1})".format(config.submatrix_rows, self.n_cols))
        self.n_submatrices = (self.n_rows * self.n_cols /
                              (config.submatrix_rows * config.submatrix_cols))
        log("n_submatrices: %d", self.n_submatrices)

        with tf.device("/cpu:0"):
            # ===== CREATE VARIABLES ======
            # Get input
            global_row, global_col, count = count_matrix_input(
                count_matrix_files, config.submatrix_rows,
                config.submatrix_cols)

            # Embeddings
            self.row_embedding = embeddings_with_init(
                embedding_dim=config.embedding_size,
                vocab_size=self.n_rows,
                name="row_embedding")
            self.col_embedding = embeddings_with_init(
                embedding_dim=config.embedding_size,
                vocab_size=self.n_cols,
                name="col_embedding")
            tf.summary.histogram("row_emb", self.row_embedding)
            tf.summary.histogram("col_emb", self.col_embedding)

            matrix_log_sum = math.log(np.sum(row_sums) + 1)
            row_bias_init = [math.log(x + 1) for x in row_sums]
            col_bias_init = [math.log(x + 1) for x in col_sums]
            self.row_bias = tf.Variable(
                row_bias_init, trainable=config.trainable_bias)
            self.col_bias = tf.Variable(
                col_bias_init, trainable=config.trainable_bias)
            tf.summary.histogram("row_bias", self.row_bias)
            tf.summary.histogram("col_bias", self.col_bias)

            # Add optimizer
            l2_losses = []
            sigmoid_losses = []
            self.global_step = tf.Variable(0, name="global_step")
            learning_rate = tf.Variable(config.learning_rate,
                                        name="learning_rate")
            opt = getattr(tf.train, FLAGS.optimizer + "Optimizer")(
                learning_rate)
            tf.summary.scalar("learning_rate", learning_rate)

            all_grads = []

        devices = ["/gpu:%d" % i for i in range(FLAGS.num_gpus)] \
            if FLAGS.num_gpus > 0 else get_available_gpus()
        self.devices_number = len(devices)
        if not self.devices_number:
            devices = ["/cpu:0"]
            self.devices_number = 1
        for dev in devices:
            with tf.device(dev):
                with tf.name_scope(dev[1:].replace(":", "_")):
                    # ===== CREATE GRAPH =====
                    # Fetch embeddings.
                    selected_row_embedding = tf.nn.embedding_lookup(
                        self.row_embedding, global_row)
                    selected_col_embedding = tf.nn.embedding_lookup(
                        self.col_embedding, global_col)

                    # Fetch biases.
                    selected_row_bias = tf.nn.embedding_lookup(
                        [self.row_bias], global_row)
                    selected_col_bias = tf.nn.embedding_lookup(
                        [self.col_bias], global_col)

                    # Multiply the row and column embeddings to generate
                    # predictions.
                    predictions = tf.matmul(
                        selected_row_embedding, selected_col_embedding,
                        transpose_b=True)

                    # These binary masks separate zero from non-zero values.
                    count_is_nonzero = tf.to_float(tf.cast(count, tf.bool))
                    count_is_zero = 1 - count_is_nonzero

                    objectives = count_is_nonzero * tf.log(count + 1e-30)
                    objectives -= tf.reshape(
                        selected_row_bias, [config.submatrix_rows, 1])
                    objectives -= selected_col_bias
                    objectives += matrix_log_sum

                    err = predictions - objectives

                    # The confidence function scales the L2 loss based on
                    # the raw co-occurrence count.
                    l2_confidence = (
                        config.confidence_base +
                        config.confidence_scale * tf.pow(
                            count, config.confidence_exponent))

                    l2_loss = config.loss_multiplier * tf.reduce_sum(
                        0.5 * l2_confidence * err * err * count_is_nonzero)
                    l2_losses.append(tf.expand_dims(l2_loss, 0))

                    sigmoid_loss = config.loss_multiplier * tf.reduce_sum(
                        tf.nn.softplus(err) * count_is_zero)
                    sigmoid_losses.append(tf.expand_dims(sigmoid_loss, 0))

                    loss = l2_loss + sigmoid_loss
                    grads = opt.compute_gradients(loss)
                    all_grads.append(grads)

        with tf.device("/cpu:0"):
            # ===== MERGE LOSSES =====
            l2_loss = tf.reduce_mean(tf.concat(axis=0, values=l2_losses), 0,
                                     name="l2_loss")
            sigmoid_loss = tf.reduce_mean(
                tf.concat(axis=0, values=sigmoid_losses), 0,
                name="sigmoid_loss")
            overall_loss = l2_loss + sigmoid_loss
            average = tf.train.ExponentialMovingAverage(0.999)
            loss_average_op = average.apply(
                (overall_loss, l2_loss, sigmoid_loss))
            self.loss = average.average(overall_loss)
            tf.summary.scalar("overall_loss", self.loss)
            tf.summary.scalar("l2_loss", average.average(l2_loss))
            tf.summary.scalar("sigmoid_loss", average.average(sigmoid_loss))

            # Apply the gradients to adjust the shared variables.
            apply_gradient_ops = []
            for grads in all_grads:
                apply_gradient_ops.append(opt.apply_gradients(
                    grads, global_step=self.global_step))

            self.train_op = tf.group(loss_average_op, *apply_gradient_ops)
            self.saver = tf.train.Saver(sharded=True)

    def initialize_summary(self, sess):
        log("creating TensorBoard stuff...")
        self.summary = tf.summary.merge_all()
        self.writer = tf.summary.FileWriter(FLAGS.logs, sess.graph)
        projector_config = \
            tf.contrib.tensorboard.plugins.projector.ProjectorConfig()
        embedding_config = projector_config.embeddings.add()
        length = min(10000, self.n_rows, self.n_cols)
        self.embedding10k = tf.Variable(
            tf.zeros((length, self._config.embedding_size)),
            name="top10k_embedding")
        embedding_config.tensor_name = self.embedding10k.name
        embedding_config.metadata_path = os.path.join(
            self._config.input_base_path, "row_vocab.txt")
        tf.contrib.tensorboard.plugins.projector.visualize_embeddings(
            self.writer, projector_config)
        self.saver = tf.train.Saver((self.embedding10k,), max_to_keep=1)

    def write_summary(self, sess):
        log("writing the summary...")
        length = min(10000, self.n_rows, self.n_cols)
        assignment = self.embedding10k.assign(
            (self.row_embedding[:length] + self.col_embedding[:length]) / 2)
        summary, _, global_step = sess.run(
            (self.summary, assignment, self.global_step))
        self.writer.add_summary(summary, global_step)
        self.saver.save(
            sess, os.path.join(FLAGS.logs, "embeddings10k.checkpoint"),
            global_step)


def main(_):
    tf.logging.set_verbosity(tf.logging.INFO)
    start_time = time.time()

    omitted = {"handler", "command"}

    log("Swivel parameters:\n" + "\n".join(
        "\t{:20} {}".format(key, value) for key, value in
        sorted(FLAGS.__dict__.items()) if key not in omitted))
    # Create the output path.  If this fails, it really ought to fail now. :)
    if not os.path.isdir(FLAGS.output_base_path):
        os.makedirs(FLAGS.output_base_path)

    # Create and run model
    with tf.Graph().as_default():
        log("creating the model...")
        model = SwivelModel(FLAGS)

        # Create a session for running Ops on the Graph.
        gpu_opts = {}
        if FLAGS.per_process_gpu_memory_fraction > 0:
            gpu_opts["per_process_gpu_memory_fraction"] = \
                FLAGS.per_process_gpu_memory_fraction
        else:
            gpu_opts["allow_growth"] = True
        gpu_options = tf.GPUOptions(**gpu_opts)
        sess = tf.Session(config=tf.ConfigProto(gpu_options=gpu_options))
        if FLAGS.logs:
            model.initialize_summary(sess)

        # Run the Op to initialize the variables.
        log("initializing the variables...")
        sess.run(tf.global_variables_initializer())

        # Start feeding input
        log("starting the input threads...")
        coord = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess=sess, coord=coord)

        # Calculate how many steps each thread should run
        n_total_steps = int(FLAGS.num_epochs * model.n_rows * model.n_cols) / (
            FLAGS.submatrix_rows * FLAGS.submatrix_cols)
        n_steps_per_thread = n_total_steps / (
            FLAGS.num_concurrent_steps * model.devices_number)
        n_submatrices_to_train = model.n_submatrices * FLAGS.num_epochs
        t0 = [time.time()]
        n_steps_between_status_updates = 100
        n_steps_between_summary_updates = 10000
        status_i = [0, 0]
        status_lock = threading.Lock()
        msg = ("%%%dd/%%d submatrices trained (%%.1f%%%%), "
               "%%5.1f submatrices/sec | loss %%f") % \
            len(str(n_submatrices_to_train))

        def TrainingFn():
            for _ in range(int(n_steps_per_thread)):
                _, global_step, loss = sess.run((
                    model.train_op, model.global_step, model.loss))

                show_status = False
                update_summary = False
                with status_lock:
                    new_i = global_step // n_steps_between_status_updates
                    if new_i > status_i[0]:
                        status_i[0] = new_i
                        show_status = True
                    new_i = global_step // n_steps_between_summary_updates
                    if new_i > status_i[1]:
                        status_i[1] = new_i
                        update_summary = True
                if show_status:
                    elapsed = float(time.time() - t0[0])
                    log(msg, global_step, n_submatrices_to_train,
                        100.0 * global_step / n_submatrices_to_train,
                        n_steps_between_status_updates / elapsed, loss)
                    t0[0] = time.time()
            if update_summary and FLAGS.logs:
                model.write_summary(sess)

    # Start training threads
    train_threads = []
    for _ in range(FLAGS.num_concurrent_steps):
        t = threading.Thread(target=TrainingFn)
        train_threads.append(t)
        t.start()

    # Wait for threads to finish.
    for t in train_threads:
        t.join()

    coord.request_stop()
    coord.join(threads)

    # Write out vectors
    write_embeddings_to_disk(FLAGS, model, sess)

    # Shutdown
    sess.close()
    log("Elapsed: %s", time.time() - start_time)


if __name__ == "__main__":
    tf.app.run()



import logging
from uuid import uuid4

from sourced.ml.extractors import create_extractors_from_args
from sourced.ml.transformers import UastDeserializer, BagFeatures2TermFreq, Uast2BagFeatures, \
    TFIDF, Cacher, Indexer, UastRow2Document, BOWWriter, Moder, create_uast_source, \
    Repartitioner, PartitionSelector, Transformer, Distinct, Collector, FieldsSelector
from sourced.ml.utils import EngineConstants
from sourced.ml.utils.engine import pipeline_graph, pause
from sourced.ml.utils.docfreq import create_or_load_ordered_df
from sourced.ml.utils.quant import create_or_apply_quant
from sourced.ml.models import DocumentFrequencies


@pause
def repos2bow_template(args, cache_hook: Transformer=None,
                       save_hook: Transformer=None):

    log = logging.getLogger("repos2bow")
    extractors = create_extractors_from_args(args)
    session_name = "repos2bow-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name)
    log.info("Loading the document index from %s ...", args.cached_index_path)
    docfreq = DocumentFrequencies().load(source=args.cached_index_path)
    document_index = {key: int(val) for (key, val) in docfreq}

    try:
        if args.quant is not None:
            create_or_apply_quant(args.quant, extractors, None)
        df_model = create_or_load_ordered_df(args, None, None)
    except ValueError:
        return 1
    ec = EngineConstants.Columns

    if args.mode == Moder.Options.repo:
        def keymap(r):
            return r[ec.RepositoryId]
    else:
        def keymap(r):
            return r[ec.RepositoryId] + UastRow2Document.REPO_PATH_SEP + \
                r[ec.Path] + UastRow2Document.PATH_BLOB_SEP + r[ec.BlobId]

    log.info("Caching UASTs to disk after partitioning by document ...")
    start_point = start_point.link(Moder(args.mode)) \
        .link(Repartitioner.maybe(args.num_iterations, keymap=keymap)) \
        .link(Cacher.maybe("DISK_ONLY"))
    for num_part in range(args.num_iterations):
        log.info("Running job %s of %s", num_part + 1, args.num_iterations)
        selected_part = start_point \
            .link(PartitionSelector(num_part))  \
            .link(Repartitioner.maybe(args.partitions, args.shuffle)) \
            .link(Cacher.maybe(args.persist))
        if cache_hook is not None:
            selected_part.link(cache_hook()).execute()
        uast_extractor = selected_part \
            .link(UastRow2Document()) \
            .link(Cacher.maybe(args.persist))
        log.info("Collecting distinct documents ...")
        documents = uast_extractor \
            .link(FieldsSelector([Uast2BagFeatures.Columns.document])) \
            .link(Distinct()) \
            .link(Collector()) \
            .execute()
        selected_part.unpersist()
        documents = set(row.document for row in documents)
        reduced_doc_index = {
            key: document_index[key] for key in document_index if key in documents}
        document_indexer = Indexer(Uast2BagFeatures.Columns.document, reduced_doc_index)
        log.info("Processing %s distinct documents", len(documents))
        bags = uast_extractor \
            .link(UastDeserializer()) \
            .link(Uast2BagFeatures(*extractors)) \
            .link(BagFeatures2TermFreq()) \
            .link(Cacher.maybe(args.persist))
        log.info("Extracting UASTs and collecting distinct tokens ...")
        tokens = bags \
            .link(FieldsSelector([Uast2BagFeatures.Columns.token])) \
            .link(Distinct()) \
            .link(Collector()) \
            .execute()
        uast_extractor.unpersist()
        tokens = set(row.token for row in tokens)
        reduced_token_freq = {key: df_model[key] for key in df_model.df if key in tokens}
        reduced_token_index = {key: df_model.order[key] for key in df_model.df if key in tokens}
        log.info("Processing %s distinct tokens", len(reduced_token_freq))
        log.info("Indexing by document and token ...")
        bags_writer = bags \
            .link(TFIDF(reduced_token_freq, df_model.docs, root.session.sparkContext)) \
            .link(document_indexer) \
            .link(Indexer(Uast2BagFeatures.Columns.token, reduced_token_index))
        if save_hook is not None:
            bags_writer = bags_writer \
                .link(Repartitioner.maybe(args.partitions, args.shuffle)) \
                .link(save_hook())
        bow = args.bow.split(".asdf")[0] + "_" + str(num_part + 1) + ".asdf"
        bags_writer \
            .link(Repartitioner.maybe(
                args.partitions, keymap=lambda x: x[Uast2BagFeatures.Columns.document])) \
            .link(BOWWriter(document_indexer, df_model, bow, args.batch)) \
            .execute()
        bags.unpersist()
    pipeline_graph(args, log, root)


def repos2bow(args):
    return repos2bow_template(args)


def repos2bow_index_template(args):
    log = logging.getLogger("repos2bow_index")
    extractors = create_extractors_from_args(args)
    session_name = "repos2bow_index_features-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name)
    uast_extractor = start_point.link(Moder(args.mode)) \
        .link(Repartitioner.maybe(args.partitions, args.shuffle)) \
        .link(UastRow2Document()) \
        .link(Cacher.maybe(args.persist))
    log.info("Extracting UASTs and indexing documents ...")
    document_indexer = Indexer(Uast2BagFeatures.Columns.document)
    uast_extractor.link(document_indexer).execute()
    document_indexer.save_index(args.cached_index_path)
    ndocs = len(document_indexer)
    log.info("Number of documents: %d", ndocs)
    uast_extractor = uast_extractor.link(UastDeserializer())
    if args.quant:
        create_or_apply_quant(args.quant, extractors, uast_extractor)
    if args.docfreq_out:
        create_or_load_ordered_df(args, ndocs, uast_extractor.link(Uast2BagFeatures(*extractors)))
    pipeline_graph(args, log, root)


def repos2bow_index(args):
    return repos2bow_index_template(args)



# flake8: noqa
from sourced.ml.algorithms.id_splitter.features import prepare_features, read_identifiers
from sourced.ml.algorithms.id_splitter.nn_model import build_rnn, build_cnn, prepare_devices, \
    register_metric, METRICS
from sourced.ml.algorithms.id_splitter.pipeline import build_schedule, make_lr_scheduler, \
    prepare_callbacks, build_train_generator, create_generator_params, str2ints, config_keras, \
    report, set_random_seed, binarize



import codecs
from collections import defaultdict
import os

from sourced.ml.algorithms import UastIds2Bag, uast2sequence
from sourced.ml.extractors import BagsExtractor, register_extractor
from sourced.ml.utils import LITERAL


class HashedTokenParser:
    def process_token(self, token):
        yield codecs.encode((hash(token) & 0xffffffffffffffff).to_bytes(8, "little"),
                            "hex_codec").decode()


class Literals2Bag(UastIds2Bag):
    """
    Converts a UAST to a bag-of-literals.
    """

    XPATH = "//*[@roleLiteral]"

    def __init__(self, token2index=None, token_parser=None):
        """
        :param token2index: The mapping from tokens to bag keys. If None, no mapping is performed.
        :param token_parser: Specify token parser if you want to use a custom one. \
            :class:'TokenParser' is used if it is not specified.
        """
        token_parser = HashedTokenParser() if token_parser is None else token_parser
        super().__init__(token2index, token_parser)

    def __call__(self, uast):
        """
        HOTFIX for https://github.com/bblfsh/client-python/issues/92
        Converts a UAST to a weighed bag-of-literals. The weights are literals frequencies.
        The tokens are preprocessed by _token_parser.
        Overwrite __call__ to avoid issues with `bblfsh.filter`.

        :param uast: The UAST root node.
        :return: bag
        """
        nodes = [node for node in uast2sequence(uast) if LITERAL in node.roles]
        bag = defaultdict(int)
        for node in nodes:
            for sub in self._token_parser.process_token(node.token):
                try:
                    bag[self._token2index[sub]] += 1
                except KeyError:
                    continue
        return bag


@register_extractor
class LiteralsBagExtractor(BagsExtractor):
    NAME = "lit"
    NAMESPACE = "l."
    OPTS = BagsExtractor.OPTS.copy()

    def __init__(self, docfreq_threshold=None, **kwargs):
        super().__init__(docfreq_threshold, **kwargs)
        self.id2bag = Literals2Bag(None, HashedTokenParser())

    def uast_to_bag(self, uast):
        if os.getenv("PYTHONHASHSEED", "random") == "random":
            raise RuntimeError("PYTHONHASHSEED must be set")
        return self.id2bag(uast)



from modelforge import register_model, Model, split_strings, merge_strings


@register_model
class Id2Vec(Model):
    """
    id2vec model - source code identifier embeddings.
    """
    NAME = "id2vec"

    def construct(self, embeddings, tokens):
        self._embeddings = embeddings
        self._tokens = tokens
        self._log.info("Building the token index...")
        self._token2index = {w: i for i, w in enumerate(self._tokens)}
        return self

    def _load_tree(self, tree):
        self.construct(embeddings=tree["embeddings"].copy(),
                       tokens=split_strings(tree["tokens"]))

    def dump(self):
        return """Shape: %s
First 10 words: %s""" % (
            self.embeddings.shape, self.tokens[:10])

    @property
    def embeddings(self):
        """
        :class:`numpy.ndarray` with the embeddings of shape
        (N tokens x embedding dims).
        """
        return self._embeddings

    @property
    def tokens(self):
        """
        List with the processed source code identifiers.
        """
        return self._tokens

    def items(self):
        """
        Returns the tuples belonging to token -> index mapping.
        """
        return self._token2index.items()

    def __getitem__(self, item):
        """
        Returns the index of the specified processed source code identifier.
        """
        return self._token2index[item]

    def __len__(self):
        """
        Returns the number of tokens in the model.
        """
        return len(self._tokens)

    def _generate_tree(self):
        return {"embeddings": self.embeddings, "tokens": merge_strings(self.tokens)}



import string
from typing import Callable, List, Tuple, Union
import warnings

import numpy
import keras
from keras import backend as kbackend
from keras.layers import BatchNormalization, Concatenate, Conv1D, Dense, Embedding, \
    Input, TimeDistributed
from keras.models import Model
try:
    import tensorflow as tf
except ImportError:
    warnings.warn("Tensorflow is not installed, dependent functionality is unavailable.")


LOSS = "binary_crossentropy"
METRICS = ["accuracy"]
# Number of unique characters and dimension of the embedding layer
NUM_CHARS = len(string.ascii_lowercase)


def register_metric(metric: Union[str, Callable]) -> Union[str, Callable]:
    """
    Decorator function to register the metrics in the METRICS constant.

    :param metrics: name of the tensorflow metric or custom function metric.
    :return: the metric.
    """
    assert isinstance(metric, str) or callable(metric)
    METRICS.append(metric)
    return metric


def prepare_devices(devices: str) -> Tuple[str]:
    """
    Extract devices from arguments.

    :param devices: devices to use passed as one string argument.
    :return: splitted devices.
    """
    devices = devices.split(",")
    if len(devices) == 2:
        dev0, dev1 = ("/gpu:" + dev for dev in devices)
    elif len(devices) == 1:
        if int(devices[0]) != -1:
            dev0 = dev1 = "/gpu:" + devices[0]
        else:
            dev0 = dev1 = "/cpu:0"
    else:
        raise ValueError("Expected 1 or 2 devices but got %d from the devices argument %s" %
                         (len(devices), devices))
    return dev0, dev1


def prepare_input_emb(maxlen: int) -> Tuple[tf.Tensor]:
    """
    Builds character embeddings, a dense representation of characters to feed the RNN with.

    :param maxlen: maximum length of the input sequence.
    :param n_uniq: number of unique characters.
    :return: input and one-hot character embedding layer.
    """
    char_seq = Input((maxlen,))
    emb = Embedding(input_dim=NUM_CHARS + 1, output_dim=NUM_CHARS + 1, input_length=maxlen,
                    mask_zero=False, weights=[numpy.eye(NUM_CHARS + 1)], trainable=False)(char_seq)
    return char_seq, emb


def add_output_layer(hidden_layer: tf.Tensor) -> keras.layers.wrappers.TimeDistributed:
    """
    Applies a Dense layer to each of the timestamps of a hidden layer, independently.
    The output layer has 1 sigmoid per character which predicts if there is a space or not
    before the character.

    :param input_layer: hidden layer before the output layer.
    :return: output layer.
    """
    norm_input = BatchNormalization()(hidden_layer)
    return TimeDistributed(Dense(1, activation="sigmoid"))(norm_input)


def add_rnn(X: tf.Tensor, units: int, rnn_layer: str, dev0: str="/gpu:0",
            dev1: str="/gpu:1") -> tf.Tensor:
    """
    Adds a bidirectional RNN layer with the specified parameters.

    :param X: input layer.
    :param units: number of neurons in the output layer.
    :param rnn_layer: type of cell in the RNN.
    :param dev0: device that will be used as forward pass of RNN and concatenation.
    :param dev1: device that will be used as backward pass.
    :return: output bidirectional RNN layer.
    """
    # select the type of RNN layer
    rnn_layer = getattr(keras.layers, rnn_layer)

    # add the forward & backward RNN
    with tf.device(dev0):
        forward = rnn_layer(units=units, return_sequences=True)(X)
    with tf.device(dev1):
        backward = rnn_layer(units=units, return_sequences=True, go_backwards=True)(X)

    # concatenate
    with tf.device(dev1):
        bidi = Concatenate(axis=-1)([forward, backward])
    return bidi


def build_rnn(maxlen: int, units: int, stack: int, optimizer: str, dev0: str,
              dev1: str, rnn_layer: str) -> keras.engine.training.Model:
    """
    Builds a RNN model with the parameters specified as arguments.

    :param maxlen: maximum length of the input sequence.
    :param units: number of neurons or dimensionality of the output RNN.
    :param stack: number of RNN layers to stack.
    :param optimizer: algorithm to use as an optimizer for the RNN.
    :param rnn_layer: recurrent layer type to use.
    :param dev0: first device to use when running specific operations.
    :param dev1: second device to use when running specific operations.
    :return: compiled RNN model.
    """
    # prepare the model
    with tf.device(dev0):
        char_seq, hidden_layer = prepare_input_emb(maxlen)

        # stack the BiDi-RNN layers
        for _ in range(stack):
            hidden_layer = add_rnn(hidden_layer, units=units, rnn_layer=rnn_layer,
                                   dev0=dev0, dev1=dev1)
        output = add_output_layer(hidden_layer)

    # compile the model
    model = Model(inputs=char_seq, outputs=output)
    model.compile(optimizer=optimizer, loss=LOSS, metrics=METRICS)
    return model


def add_conv(X: tf.Tensor, filters: List[int], kernel_sizes: List[int],
             output_n_filters: int) -> tf.Tensor:
    """
    Builds a single convolutional layer.

    :param X: input layer.
    :param filters: number of output filters in the convolution.
    :param kernel_sizes: list of lengths of the 1D convolution window.
    :param output_n_filters: number of 1D output filters.
    :return: output layer.
    """
    # normalize the input
    X = BatchNormalization()(X)

    # add convolutions
    convs = []

    for n_filters, kernel_size in zip(filters, kernel_sizes):
        conv = Conv1D(filters=n_filters, kernel_size=kernel_size, padding="same",
                      activation="relu")
        convs.append(conv(X))

    # concatenate all convolutions
    conc = Concatenate(axis=-1)(convs)
    conc = BatchNormalization()(conc)

    # dimensionality reduction
    conv = Conv1D(filters=output_n_filters, kernel_size=1, padding="same", activation="relu")
    return conv(conc)


def build_cnn(maxlen: int, filters: List[int], output_n_filters: int, stack: int,
              kernel_sizes: List[int], optimizer: str, device: str) -> keras.engine.training.Model:
    """
    Builds a CNN model with the parameters specified as arguments.

    :param maxlen: maximum length of the input sequence.
    :param filters: number of output filters in the convolution.
    :param output_n_filters: number of 1d output filters.
    :param stack: number of CNN layers to stack.
    :param kernel_sizes: list of lengths of the 1D convolution window.
    :param optimizer: algorithm to use as an optimizer for the CNN.
    :param device: device to use when running specific operations.
    :return: compiled CNN model.
    """
    # prepare the model
    with tf.device(device):
        char_seq, hidden_layer = prepare_input_emb(maxlen)

        # stack the CNN layers
        for _ in range(stack):
            hidden_layer = add_conv(hidden_layer, filters=filters, kernel_sizes=kernel_sizes,
                                    output_n_filters=output_n_filters)
        output = add_output_layer(hidden_layer)

    # compile the model
    model = Model(inputs=char_seq, outputs=output)
    model.compile(optimizer=optimizer, loss=LOSS, metrics=METRICS)
    return model


@register_metric
def precision(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    """
    Computes the precision, a metric for multi-label classification of
    how many selected items are relevant.

    :param y_true: tensor of true labels.
    :param y_pred: tensor of predicted labels.
    :return: a tensor batch-wise average of precision.
    """
    true_positives = kbackend.sum(kbackend.round(kbackend.clip(y_true * y_pred, 0, 1)))
    predicted_positives = kbackend.sum(kbackend.round(kbackend.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + kbackend.epsilon())
    return precision


@register_metric
def recall(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    """
    Computes the recall, a metric for multi-label classification of
    how many relevant items are selected.

    :param y_true: tensor of true labels.
    :param y_pred: tensor of predicted labels.
    :return: a tensor batch-wise average of recall.
    """
    true_positives = kbackend.sum(kbackend.round(kbackend.clip(y_true * y_pred, 0, 1)))
    possible_positives = kbackend.sum(kbackend.round(kbackend.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + kbackend.epsilon())
    return recall


@register_metric
def f1score(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    """
    Computes the F1 score,  the harmonic average of precision and recall.

    :param y_true: tensor of true labels.
    :param y_pred: tensor of predicted labels.
    :return: a tensor batch-wise average of F1 score.
    """
    prec = precision(y_true, y_pred)
    rec = recall(y_true, y_pred)
    return 2 * prec * rec / (prec + rec + kbackend.epsilon())



from glob import glob
import operator
import os

import humanize
import numpy
from pyspark import RDD
from scipy.sparse import csr_matrix

from sourced.ml.models import OrderedDocumentFrequencies, BOW
from sourced.ml.transformers import Indexer, Uast2BagFeatures
from sourced.ml.transformers.transformer import Transformer


class BOWWriter(Transformer):
    DEFAULT_CHUNK_SIZE = 2 * 1000 * 1000 * 1000

    def __init__(self, document_indexer: Indexer, df: OrderedDocumentFrequencies,
                 filename: str, chunk_size: int=DEFAULT_CHUNK_SIZE, **kwargs):
        super().__init__(**kwargs)
        self.document_indexer = document_indexer
        self.df = df
        self.filename = filename
        self.chunk_size = chunk_size

    def __getstate__(self):
        state = super().__getstate__()
        del state["document_indexer"]
        del state["token_indexer"]
        del state["df"]
        return state

    def __call__(self, head: RDD):
        c = Uast2BagFeatures.Columns
        self._log.info("Estimating the average bag size...")
        avglen = head \
            .map(lambda x: (x[c.document], 1)) \
            .reduceByKey(operator.add) \
            .map(lambda x: x[1])
        if self.explained:
            self._log.info("toDebugString():\n%s", avglen.toDebugString().decode())
        avglen = avglen.mean()
        self._log.info("Result: %.0f", avglen)
        avgdocnamelen = numpy.mean([len(v) for v in self.document_indexer.value_to_index])
        nparts = int(numpy.ceil(
            len(self.document_indexer) * (avglen * (4 + 4) + avgdocnamelen) / self.chunk_size))
        self._log.info("Estimated number of partitions: %d", nparts)
        doc_index_to_name = {
            index: name for name, index in self.document_indexer.value_to_index.items()}
        tokens = self.df.tokens()
        it = head \
            .map(lambda x: (x[c.document], (x[c.token], x[c.value]))) \
            .groupByKey() \
            .repartition(nparts) \
            .glom() \
            .toLocalIterator()
        if self.explained:
            self._log.info("toDebugString():\n%s", it.toDebugString().decode())
        ndocs = 0
        self._log.info("Writing files to %s", self.filename)
        for i, part in enumerate(it):
            docs = [doc_index_to_name[p[0]] for p in part]
            if not len(docs):
                self._log.info("Batch %d is empty, skipping.", i + 1)
                continue
            size = sum(len(p[1]) for p in part)
            data = numpy.zeros(size, dtype=numpy.float32)
            indices = numpy.zeros(size, dtype=numpy.int32)
            indptr = numpy.zeros(len(docs) + 1, dtype=numpy.int32)
            pos = 0
            for pi, (_, bag) in enumerate(part):
                for tok, val in sorted(bag):
                    indices[pos] = tok
                    data[pos] = val
                    pos += 1
                indptr[pi + 1] = indptr[pi] + len(bag)
            assert pos == size
            matrix = csr_matrix((data, indices, indptr), shape=(len(docs), len(tokens)))
            filename = self.get_bow_file_name(self.filename, i)
            BOW() \
                .construct(docs, tokens, matrix) \
                .save(filename, deps=(self.df,))
            self._log.info("%d -> %s with %d documents, %d nnz (%s)",
                           i + 1, filename, len(docs), size,
                           humanize.naturalsize(os.path.getsize(filename)))
            ndocs += len(docs)
        self._log.info("Final number of documents: %d", ndocs)

    def get_bow_file_name(self, base: str, index: int):
        parent, full_name = os.path.split(base)
        name, ext = os.path.splitext(full_name)
        return os.path.join(parent, "%s_%03d%s" % (name, index + 1, ext))


class BOWLoader:
    def __init__(self, glob_path):
        self.files = glob(glob_path)

    def __len__(self):
        return len(self.files)

    def __bool__(self):
        return bool(self.files)

    def __iter__(self):
        return (BOW().load(path) for path in self.files)



import logging
from uuid import uuid4

from sourced.ml.extractors import IdentifierDistance
from sourced.ml.transformers import UastDeserializer, Uast2BagFeatures, UastRow2Document, \
    CsvSaver, create_uast_source
from sourced.ml.transformers.basic import Rower
from sourced.ml.utils.engine import pipeline_graph, pause


@pause
def repos2id_distance(args):
    log = logging.getLogger("repos2roles_and_ids")
    extractor = IdentifierDistance(args.split, args.type, args.max_distance)
    session_name = "repos2roles_and_ids-%s" % uuid4()
    root, start_point = create_uast_source(args, session_name)

    start_point \
        .link(UastRow2Document()) \
        .link(UastDeserializer()) \
        .link(Uast2BagFeatures(extractor)) \
        .link(Rower(lambda x: dict(identifier1=x[0][0][0],
                                   identifier2=x[0][0][1],
                                   distance=x[1]))) \
        .link(CsvSaver(args.output)) \
        .execute()
    pipeline_graph(args, log, root)



import logging
from uuid import uuid4

from modelforge.progress_bar import progress_bar
import numpy as np
from scipy.sparse import coo_matrix

from sourced.ml.cmd.args import handle_input_arg
from sourced.ml.extractors.helpers import filter_kwargs
from sourced.ml.models import OrderedDocumentFrequencies, Cooccurrences
from sourced.ml.transformers import CooccModelSaver
from sourced.ml.utils.engine import pause
from sourced.ml.utils.spark import create_spark


MAX_INT32 = 2**31 - 1


@pause
def merge_coocc(args):
    log = logging.getLogger("merge_coocc")
    log.setLevel(args.log_level)
    filepaths = list(handle_input_arg(args.input, log))
    log.info("Will merge %d files", len(filepaths))
    df = OrderedDocumentFrequencies().load(args.docfreq)
    if args.no_spark:
        merge_coocc_no_spark(df, filepaths, log, args)
    else:
        merge_coocc_spark(df, filepaths, log, args)


def load_and_check(filepaths: list, log: logging.Logger):
    """
    Load Cooccurrences models from filepaths list and perform simple check:
    1. If model contains values more than MAX_INT32 we saturate.
    2. If model contains negative values we consider it as corrupted, report and skip.
    """
    for path in progress_bar(filepaths, log):
        coocc = Cooccurrences().load(path)
        negative_values = np.where(coocc.matrix.data < 0)
        if negative_values[0].size > 0:
            log.warning("Model %s is corrupted and will be skipped. "
                        "It contains negative elements.", path)
            continue
        too_big_values = np.where(coocc.matrix.data > MAX_INT32)
        if too_big_values[0].size > 0:
            log.warning("Model %s contains elements with values more than MAX_INT32. "
                        "They will be saturated to MAX_INT32", path)
            coocc.matrix.data[too_big_values] = MAX_INT32
        yield path, coocc


def merge_coocc_spark(df, filepaths, log, args):
    session_name = "merge_coocc-%s" % uuid4()
    session = create_spark(session_name, **filter_kwargs(args.__dict__, create_spark))
    spark_context = session.sparkContext
    global_index = spark_context.broadcast(df.order)

    coocc_rdds = []

    def local_to_global(local_index):
        """
        Converts token index of co-occurrence matrix to the common index.
        For example index, 5 correspond to `get` token for a current model.
        And `get` have index 7 in the result.
        So we convert 5 to `get` via tokens list and `get` to 7 via global_index mapping.
        If global_index do not have `get` token, it returns -1.
        """
        return global_index.value.get(tokens.value[local_index], -1)

    for path, coocc in load_and_check(filepaths, log):
        rdd = coocc.matrix_to_rdd(spark_context)  # rdd structure: ((row, col), weight)
        log.info("Broadcasting tokens order for %s model...", path)
        tokens = spark_context.broadcast(coocc.tokens)
        coocc_rdds.append(
            rdd.map(lambda row: ((local_to_global(row[0][0]),
                                  local_to_global(row[0][1])),
                                 np.uint32(row[1])))
               .filter(lambda row: row[0][0] >= 0))

    log.info("Calculating the union of cooccurrence matrices...")
    rdd = spark_context \
        .union(coocc_rdds) \
        .reduceByKey(lambda x, y: min(MAX_INT32, x + y))
    CooccModelSaver(args.output, df)(rdd)


def merge_coocc_no_spark(df, filepaths, log, args):
    """
    Algorithm explanation:

    1. Although we store result in uint32, we actually never have elements greater than MAX_INT32
    2. We assume that both result and the summed matrix do not have elements greater than MAX_INT32
    3. As soon as we have a value bigger than MAX_INT32 after summing, we saturate
    4. Thus we lose 2x data range but do not allocate any additional memory and it works faster
       than MAX_UINT32 checks
    5. Only ? elements saturate in PGA so this is fine

    """
    # TODO(zurk): recheck the number of saturated elements.
    log.info("Merging cooccurrences without using PySpark")
    shape = (len(df) + 1,) * 2
    result = coo_matrix(shape, dtype=np.uint32)
    for path, coocc in load_and_check(filepaths, log):
        coocc._matrix = coo_matrix(coocc._matrix)
        index = [df.order.get(x, len(df)) for x in coocc.tokens]
        rows = [index[x] for x in coocc.matrix.row]
        cols = [index[x] for x in coocc.matrix.col]
        result += coo_matrix(
            (coocc.matrix.data, (rows, cols)), shape=shape, dtype=np.uint32)
        indx_overflow = np.where(result.data > MAX_INT32)
        if indx_overflow[0].size > 0:
            log.warning("Overflow in %d elements. They will be saturated to MAX_INT32",
                        indx_overflow[0].size)
            result.data[indx_overflow] = MAX_INT32
    Cooccurrences() \
        .construct(df.tokens(), result[:-1, :-1]) \
        .save(args.output, (df,))



from typing import Iterable, Tuple, Union
from itertools import combinations

import bblfsh

from sourced.ml.algorithms.uast_ids_to_bag import UastIds2Bag
from sourced.ml.utils import bblfsh_roles


class Uast2IdDistance(UastIds2Bag):
    """
    Converts a UAST to a list of identifiers pair and UAST distance between.
    Distance metric must be defined in the inheritors.

    __call__ is overridden here and return list instead of bag-of-words (dist).
    """

    DEFAULT_MAX_DISTANCE = 10  # to avoid collecting all distances we skip too big ones

    def __init__(self, token2index=None, token_parser=None, max_distance=DEFAULT_MAX_DISTANCE):
        """
        :param token2index: The mapping from tokens to token key. If None, no mapping is performed.
        :param token_parser: Specify token parser if you want to use a custom one. \
            :class:'TokenParser' is used if it is not specified.
        """
        super().__init__(token2index=token2index, token_parser=token_parser)
        self.max_distance = max_distance

    def __call__(self, uast: bblfsh.Node) -> Iterable[Tuple[str, str, int]]:
        """
        Converts a UAST to a list of identifiers pair and UAST distance between.
        The tokens are preprocessed by _token_parser.

        :param uast: The UAST root node.
        :return: a list of (from identifier, to identifier) and distance pairs.
        """
        for point1, point2 in combinations(self._process_uast(uast), 2):
            if point1[0] == point2[0]:
                continue  # We do not want to calculate distance between the same identifiers
            distance = self.distance(point1, point2)
            if distance < self.max_distance:
                yield ((point1[0], point2[0]) if point1[0] > point2[0] else
                       (point2[0], point1[0])), distance

    def distance(self, point1, point2) -> Union[int, float]:
        """
        Calculate distance between two points. A point can be anything. self._process_uast returns
        list of points in the specific class.

        :return: Distance between two points.
        """
        raise NotImplemented

    def _process_uast(self, node: bblfsh.Node) -> Iterable:
        """
        Converts uast to points list. A point can be anything you need to calculate distance.
        """
        raise NotImplemented

    def _process_point(self, node, info):
        if bblfsh_roles.IDENTIFIER in node.roles and node.token:
            for sub in self._token_parser.process_token(node.token):
                try:
                    yield (self._token2index[sub], info)
                except KeyError:
                    continue


class Uast2IdTreeDistance(Uast2IdDistance):
    """
    Converts a UAST to a list of identifiers pair and UAST tree distance between.

    __call__ is overridden here and return list instead of bag-of-words (dist).
    """
    def _process_uast(self, uast: bblfsh.Node) -> Iterable:
        stack = [(uast, [])]
        while stack:
            node, ancestors = stack.pop()
            yield from self._process_point(node, ancestors)
            ancestors = list(ancestors)
            ancestors.append(node)
            stack.extend([(child, ancestors) for child in node.children])

    def distance(self, point1, point2) -> int:
        i = 0
        ancestors1 = point1[1]
        ancestors2 = point2[1]
        for i, (ancestor1, ancestor2) in enumerate(zip(ancestors1, ancestors2)):
            if ancestor1 != ancestor2:
                break
        distance = self.calc_tree_distance(i, len(ancestors1), len(ancestors2))
        return distance

    @staticmethod
    def calc_tree_distance(last_common_level, level1, level2):
        return level1 + level2 - 2 * last_common_level


class Uast2IdLineDistance(Uast2IdDistance):
    """
    Converts a UAST to a list of identifiers pair and code line distance between where applicable.

    __call__ is overridden here and return list instead of bag-of-words (dist).
    """

    def _process_uast(self, uast):
        stack = [(uast, [0, 0])]
        while stack:
            node, last_position = stack.pop()
            if node.start_position.line != 0:
                # A lot of Nodes do not have position
                # It is good heuristic to take the last Node in tree with a position.
                last_position[0] = node.start_position.line
                last_position[1] = 0
            if node.start_position.col != 0:
                last_position[1] = node.start_position.col
            yield from self._process_point(node, last_position)
            stack.extend([(child, list(last_position)) for child in node.children])

    def distance(self, point1, point2):
        return abs(point1[1][0] - point2[1][0])  # subtract line numbers



import logging

from sourced.ml.models import MergeBOW
from sourced.ml.cmd.args import handle_input_arg


def merge_bow(args):
    log = logging.getLogger("merge_bow")
    MergeBOW(args.features).convert(handle_input_arg(args.input, log), args.output)
